/***************************************************************************
 *   Copyright (C) 2006 by Borko Boskovic                                  *
 *   borko.boskovic@uni-mb.si                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

inline gboolean bishopsCannotMate(guint64 bishops){
	gboolean white = FALSE, black = FALSE;
	while(bishops){
		if(getFirstBitAndClear64(&bishops) % 2 == 0) white = TRUE;
		else black = TRUE;
		if(white && black) return FALSE;
	}
	return TRUE;
}

inline gint kingsDistance(const Board* board){
	return distance(indexOfFirstBit64(&board->king[WHITE]),
				indexOfFirstBit64(&board->king[BLACK]));
}

inline gint KBBK(const Board * board){
	gint loneKingPlayer = board->player;
	if(board->bishops[loneKingPlayer] != 0) loneKingPlayer = board->opponent;

	if(bishopsCannotMate(board->bishops[loneKingPlayer^1])) return 0;

	gint score= 11.0*PAWN_VALUE-kingsDistance(board)*(PAWN_VALUE/8.0)-
				cornerDistance(indexOfFirstBit64(&board->king[loneKingPlayer]))*
				(PAWN_VALUE/4.0);
				
	if(loneKingPlayer != board->opponent)
		score = -score;
	return score;
}

inline gint KBNK(const Board * board){
 	gint loneKingPlayer = board->player;
	if(board->bishops[loneKingPlayer] != 0) loneKingPlayer = board->opponent;
	
	gint king = indexOfFirstBit64(&board->king[loneKingPlayer]);
	gint bishop = indexOfFirstBit64(&board->bishops[loneKingPlayer^1]);
	gint f = file(bishop); 
	gint r = rank(bishop); 
	gint color = WHITE;
	gint kb_cornerDistance;
	if((r%2 == 0 && f%2 == 0) || (r%2 != 0 && f%2 != 0)) color = BLACK;
	if(color == WHITE) kb_cornerDistance = MIN(distance(king,h1),distance(king,a8));
	else kb_cornerDistance = MIN(distance(king,a1),distance(king,h8));
	int  score = 10.0*PAWN_VALUE - kb_cornerDistance*(PAWN_VALUE/4.0) - 		
				kingsDistance(board)*(PAWN_VALUE/8.0);
	if(loneKingPlayer != board->opponent) score = -score;
	return score;
}

inline gint KRK(const Board* board){
 	gint loneKingPlayer = board->player;
	if(board->rooks[loneKingPlayer] != 0) loneKingPlayer = board->opponent;
	
	gint king = indexOfFirstBit64(&board->king[loneKingPlayer]);
	gint f = file(king); 
	gint r = rank(king); 
	gint distance = MIN(MIN((7-f),f),MIN((7-r),r));
	gint score = 12.0*PAWN_VALUE - distance*(PAWN_VALUE/4.0)-kingsDistance(board)*(PAWN_VALUE/8.0);
	if(loneKingPlayer != board->opponent) score = -score;
	return score;
}

inline gint KQK(const Board* board){
 	gint loneKingPlayer = board->player;
	if(board->queens[loneKingPlayer] != 0) loneKingPlayer = board->opponent;
	
	gint king = indexOfFirstBit64(&board->king[loneKingPlayer]);
	gint f = file(king); 
	gint r = rank(king); 
	gint distance = MIN(MIN((7-f),f),MIN((7-r),(r)));
	gint score = 13.0*PAWN_VALUE - distance*(PAWN_VALUE/4.0)-kingsDistance(board)*(PAWN_VALUE/8.0);
	if(loneKingPlayer != board->opponent) score = -score;
	return score;
}

gint boardEvaluateEndGame(const Board* board){
		gint material = board->material[WHITE] - board->material[BLACK];
		gint numberOfPieces = board->numberOfPieces[WHITE] + board->numberOfPieces[BLACK];
		switch(numberOfPieces){
			case 2:{ return 0; } /* KK */
			case 3:{
				switch(material){
					case BISHOP_VALUE 	:
					case -BISHOP_VALUE	:{ return 0; }/* KBK */
					case KNIGHT_VALUE	:
					case -KNIGHT_VALUE	:{ return 0; } /* KNK */
					case ROOK_VALUE		:
					case -ROOK_VALUE	:{ return KRK(board); } /* KRK */
					case QUEEN_VALUE	:
					case -QUEEN_VALUE	:{ return KQK(board); } /* KQK */
					case PAWN_VALUE		:
					case -PAWN_VALUE	:{ return material; } /* KPK ??? */
				}
				break;
			}
			case 4:{
				switch(material){
					case(BISHOP_VALUE + BISHOP_VALUE)	:
					case(-(BISHOP_VALUE+BISHOP_VALUE))	:{ return KBBK(board); } /* KBBK */
					case(BISHOP_VALUE+KNIGHT_VALUE)		:
					case(-(BISHOP_VALUE+KNIGHT_VALUE))	:{ return KBNK(board); } /* KBNK */
					case(KNIGHT_VALUE+KNIGHT_VALUE)		:
					case(-(KNIGHT_VALUE+KNIGHT_VALUE))	:{ return 0; } /* KNNK */
					case 0								:{
						if(board->knights[WHITE] != 0) return 0; /* KNKN */
						if(board->bishops[WHITE] != 0) return 0; /* KBKB */
					}
					case(KNIGHT_VALUE - BISHOP_VALUE)	:
					case(BISHOP_VALUE - KNIGHT_VALUE)	:{ return 0; } /* KBKN */
					case(ROOK_VALUE - KNIGHT_VALUE)		:
					case(-(ROOK_VALUE - KNIGHT_VALUE))	:{ return 0; } /* KRKN */
					case(ROOK_VALUE - BISHOP_VALUE)		:
					case(-(ROOK_VALUE - BISHOP_VALUE))	:{ return 0; } /* BRKB */
				}
				break;
			}
		}
	return material;
}
