/***************************************************************************
 *   Copyright (C) 2006 by Borko Boskovic                                  *
 *   borko.boskovic@uni-mb.si                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "boardEndGameEvaluator.c"

const gint KnightPhase = 1;
const gint BishopPhase = 1;
const gint RookPhase   = 2;
const gint QueenPhase  = 4;
const gint TotalPhase =  24; /*KnightPhase * 4 + BishopPhase * 4 + RookPhase * 4 + QueenPhase * 2*/

const gint PawnValueOpening = 70;
const gint KnightValueOpening = 325;
const gint BishopValueOpening = 325;
const gint RookValueOpening = 500;
const gint QueenValueOpening = 975;

const gint PawnValueEndgame = 90;
const gint KnightValueEndgame = 325;
const gint BishopValueEndgame = 325;
const gint RookValueEndgame = 500;
const gint QueenValueEndgame = 975;

const gint PawnFileOpening = 5;

const gint KnightCenterOpening = 5;
const gint KnightCenterEndgame = 5;
const gint KnightRankOpening = 5;
const gint KnightBackRankOpening = 0;
const gint KnightTrapped = 100;

const gint BishopCenterOpening = 2;
const gint BishopCenterEndgame = 3;
const gint BishopBackRankOpening = 10;
const gint BishopDiagonalOpening = 4;
const gint BishopPairOpening = 50;
const gint BishopPairEndgame = 50;

const gint RookFileOpening = 3;

const gint QueenCenterOpening = 0;
const gint QueenCenterEndgame = 4;
const gint QueenBackRankOpening = 5;

const gint KingCenterEndgame = 12;
const gint KingFileOpening = 10;
const gint KingRankOpening = 10;
const gint KingAttackOpening = 20;

const gint PawnFile[8] = {-3, -1, +0, +1, +1, +0, -1, -3};
const gint KnightLine[8] = {-4, -2, +0, +1, +1, +0, -2, -4};
const gint KnightRank[8] = {-2, -1, +0, +1, +2, +3, +2, +1};
const gint BishopLine[8] = {-3, -1, +0, +1, +1, +0, -1, -3};
const gint RookFile[8] = {-2, -1, +0, +1, +1, +0, -1, -2};
const gint QueenLine[8] = {-3, -1, +0, +1, +1, +0, -1, -3};
const gint KingLine[8] = {-3, -1, +0, +1, +1, +0, -1, -3};
const gint KingFile[8] = {+3, +4, +2, +0, +0, +2, +4, +3};
const gint KingRank[8] = {+1, +0, -2, -3, -4, -5, -6, -7};
const gint KingAttackUnit[14] = {0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 4, 4, 0 ,0};
const gint KingAttackWeight[16] = {0, 0, 128, 192, 224, 240, 248, 252, 254, 255, 256, 256 ,256, 256, 256, 256};
const gint Bonus[8] = {0,0,0,26,77,154,256,0};

const gint TrappedBishop = 100;
const gint BlockedBishop = 50;
const gint BlockedRook = 50;

const gint KnightMobOpening = 4;
const gint KnightMobEndgame = 4;
const gint BishopMobOpening = 5;
const gint BishopMobEndgame = 5;
const gint RookMobOpening = 2;
const gint RookMobEndgame = 4;
const gint QueenMobOpening = 1;
const gint QueenMobEndgame = 2;
const gint KingMobOpening = 0;
const gint KingMobEndgame = 0;
const gint RookSemiOpenFileOpening = 10;
const gint RookSemiOpenFileEndgame = 10;
const gint RookOpenFileOpening = 20;
const gint RookOpenFileEndgame = 20;
const gint RookSemiKingFileOpening = 10;
const gint RookKingFileOpening = 20;
const gint Rook7thOpening = 20;
const gint Rook7thEndgame = 40;
const gint Queen7thOpening = 10;
const gint Queen7thEndgame = 20;

const gint CandidateOpeningMin = 5;
const gint CandidateOpeningMax = 55;
const gint CandidateEndgameMin = 10;
const gint CandidateEndgameMax = 110;

const gint KnightOutpostMatrix[2][64] = {
{	 
	0, 0, 0, 0, 0, 0, 0, 0,
     	0, 0, 0, 0, 0, 0, 0, 0,
     	0, 0, 0, 0, 0, 0, 0, 0,
     	0, 2, 5, 10, 10, 5, 2, 0,
     	0, 2, 5, 10, 10, 5, 2, 0,
     	0, 0, 4, 5, 5, 4, 0, 0,
     	0, 0, 0, 0, 0, 0, 0, 0,
     	0, 0, 0, 0, 0, 0, 0, 0
},
{
	0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 4, 5, 5, 4, 0, 0,
	0, 2, 5, 10, 10, 5, 2, 0,
	0, 2, 5, 10, 10, 5, 2, 0,
	0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0
}
};

const gint KnightUnit = 4;
const gint BishopUnit = 6;
const gint RookUnit = 7;
const gint QueenUnit = 13;

const gint StormOpening = 10;

const gint DoubledOpening = 10;
const gint DoubledEndgame = 20;

const gint IsolatedOpening = 10;
const gint IsolatedOpeningOpen = 20;
const gint IsolatedEndgame = 20;

const gint BackwardOpening = 8;
const gint BackwardOpeningOpen = 16;
const gint BackwardEndgame = 10;

const gint PassedOpeningMin = 10;
const gint PassedOpeningMax = 70;
const gint PassedEndgameMin = 20;
const gint PassedEndgameMax = 140;

const gint UnstoppablePasser = 800;
const gint FreePasser = 60;

const gint AttackerDistance = 5;
const gint DefenderDistance = 20;

struct sevalInfo{
	gint kingSafety[2];
	gint majorMaterial[2];
	gint phase;
	gint evalOpening;
	gint evalEndgame;
	guint64 passedPawn[2];
};

typedef struct sevalInfo EvalInfo;

/**
 *
 */
void boardPstInit(){
	gint i,j,k;
	for(i=0; i<14; i++){
		for(j=0; j<64; j++){
			for(k=0; k<2; k++){
				pst[i][j][k] = 0;
			}
		}
	}

	/* Pawns */
	for(i=0; i<64; i++) pst[PAWN][i][Opening] += PawnFile[file(i)] * PawnFileOpening;
	/* Center control */
	pst[PAWN][d3][Opening] += 10;
	pst[PAWN][e3][Opening] += 10;
	pst[PAWN][d4][Opening] += 20;
	pst[PAWN][e4][Opening] += 20;
	pst[PAWN][d5][Opening] += 10;
	pst[PAWN][e5][Opening] += 10;

	/* Knights */
	for(i=0; i<64; i++){
		pst[KNIGHT][i][Opening] += KnightLine[file(i)] * KnightCenterOpening;
		pst[KNIGHT][i][Opening] += KnightLine[rank(i)] * KnightCenterOpening;
		pst[KNIGHT][i][Opening] += KnightRank[rank(i)] * KnightRankOpening;
		pst[KNIGHT][i][Endgame] += KnightLine[file(i)] * KnightCenterEndgame;
		pst[KNIGHT][i][Endgame] += KnightLine[rank(i)] * KnightCenterEndgame;
	}
	/* Back rank */
	for(i=a1; i<=h1; i++) pst[KNIGHT][i][Opening] -= KnightBackRankOpening;
	/* Trapped */
	pst[KNIGHT][a8][Opening] -= KnightTrapped;
	pst[KNIGHT][h8][Opening] -= KnightTrapped;

	/* Bishops */
	for(i=0; i<64; i++){
		pst[BISHOP][i][Opening] += BishopLine[file(i)] * BishopCenterOpening;
		pst[BISHOP][i][Opening] += BishopLine[rank(i)] * BishopCenterOpening;
		pst[BISHOP][i][Endgame] += BishopLine[file(i)] * BishopCenterEndgame;
		pst[BISHOP][i][Endgame] += BishopLine[rank(i)] * BishopCenterEndgame;
	}
	/* Back rank */
	for(i=a1; i<=h1; i++) pst[BISHOP][i][Opening] -= BishopBackRankOpening;
	/* Main diagonals*/
	for(i=0; i<8; i++){
		j = getSquare(i,i);
		pst[BISHOP][j][Opening] += BishopDiagonalOpening;
		pst[BISHOP][opposite(j)][Opening] += BishopDiagonalOpening;
	}

	/* Rooks */
	for(i=0; i<64; i++) pst[ROOK][i][Opening] += RookFile[file(i)] * RookFileOpening;

	/* Queens */
	for(i=0; i<64; i++){
		pst[QUEEN][i][Opening] += QueenLine[file(i)] * QueenCenterOpening;
		pst[QUEEN][i][Opening] += QueenLine[rank(i)] * QueenCenterOpening;
		pst[QUEEN][i][Endgame] += QueenLine[file(i)] * QueenCenterEndgame;
		pst[QUEEN][i][Endgame] += QueenLine[rank(i)] * QueenCenterEndgame;
	}
	/* Back rank */
	for(i=a1; i<=h1; i++) pst[QUEEN][i][Opening] -= QueenBackRankOpening;

	/* Kings */
	for(i=0; i< 64; i++){
		pst[KING][i][Opening] += KingFile[file(i)] * KingFileOpening;
		pst[KING][i][Opening] += KingRank[rank(i)] * KingRankOpening;
		pst[KING][i][Endgame] += KingLine[file(i)] * KingCenterEndgame;
		pst[KING][i][Endgame] += KingLine[rank(i)] * KingCenterEndgame;
	}

	/* Symmetry copy for black */
	for (i = 0; i < 14; i += 2) {
		for (j = 0; j < 64; j++) {
			for (k= 0; k < 2; k++) {
				pst[i+1][j][k] = -pst[i][opposite(j)][k];
         		}
      		}
   	}

	/* Print pst */
	/*for (k= 0; k < 2; k++) {
		for (i = 0; i < 14; i ++) {
			g_print("%c:\n",pieceToChar[i]);
			for (j = 0; j < 64; j++) {
				if(j%8==0) g_print("\n");
				g_print("%d ",pst[i][j][k]);
			}
			g_print("\n");
      		}
		g_print("\n");
   	}*/
}

/**
 *
 */
void boardInitEvaluator(){
	boardPstInit();
}

/**
 *
 */
void boardMaterialInfo(const Board* board, EvalInfo* evalInfo){
	gint evalOpening=0, evalEndgame=0, kingSafety[2], majorMaterial[2];

	/* init */
	gint wp = numberOfBits64(board->pawns[WHITE]);
	gint wn = numberOfBits64(board->knights[WHITE]);
   	gint wb = numberOfBits64(board->bishops[WHITE]);
	gint wr = numberOfBits64(board->rooks[WHITE]);
	gint wq = numberOfBits64(board->queens[WHITE]);

	gint bp = numberOfBits64(board->pawns[BLACK]);
	gint bn = numberOfBits64(board->knights[BLACK]);
   	gint bb = numberOfBits64(board->bishops[BLACK]);
	gint br = numberOfBits64(board->rooks[BLACK]);
	gint bq = numberOfBits64(board->queens[BLACK]);

	/* Phase */
	gint phase = TotalPhase;

	phase -= wn * KnightPhase;
	phase -= wb * BishopPhase;
	phase -= wr * RookPhase;
	phase -= wq * QueenPhase;

	phase -= bn * KnightPhase;
	phase -= bb * BishopPhase;
	phase -= br * RookPhase;
	phase -= bq * QueenPhase;

   	if(phase < 0) phase = 0;
   	phase = (phase * 256 + (TotalPhase / 2)) / TotalPhase;

	/* King safety */
	if (bq >= 1 && bq+br+bb+bn >= 2) kingSafety[WHITE] = TRUE;
	else kingSafety[WHITE] = FALSE;

	if (wq >= 1 && wq+wr+wb+wn >= 2) kingSafety[BLACK] = TRUE;
	else kingSafety[BLACK] = FALSE;

	/* Material */
	majorMaterial[WHITE] = wn*KnightValueOpening + wb*BishopValueOpening + 
								wr*RookValueOpening + wq*QueenValueOpening;
	evalOpening += majorMaterial[WHITE] + wp*PawnValueOpening;
	majorMaterial[BLACK] = bn*KnightValueOpening + bb*BishopValueOpening + 
								br*RookValueOpening + bq*QueenValueOpening;
	evalOpening -= majorMaterial[BLACK] + bp*PawnValueOpening;


	evalEndgame += wn*KnightValueEndgame + wb*BishopValueEndgame + wr*RookValueEndgame + 
					wq*QueenValueEndgame + wp*PawnValueEndgame;
	evalEndgame -= bn*KnightValueEndgame + bb*BishopValueEndgame + br*RookValueEndgame + 
					bq*QueenValueEndgame + bp*PawnValueEndgame;

	/* Bishop pair */
	if (wb >= 2){
	       evalOpening += BishopPairOpening;
	       evalEndgame += BishopPairEndgame;
	}
	if (bb >= 2){
	       evalOpening -= BishopPairOpening;
	       evalEndgame -= BishopPairEndgame;
	}

	/* Store info */
	evalInfo->kingSafety[WHITE] = kingSafety[WHITE];
	evalInfo->kingSafety[BLACK] = kingSafety[BLACK];
	evalInfo->majorMaterial[WHITE] = majorMaterial[WHITE];
	evalInfo->majorMaterial[BLACK] = majorMaterial[BLACK];
	evalInfo->phase = phase;
	evalInfo->evalOpening = evalOpening;
	evalInfo->evalEndgame = evalEndgame;
}

/**
 *
 */
void boardEvalPatterns(const Board* board, gint* evalOpening, gint* evalEndgame){
	/* Trapped bishop (7th rank) */
	if((board->types[a7] ==  WHITE_BISHOP && board->types[b6] == BLACK_PAWN) ||
	   (board->types[b8] == WHITE_BISHOP && board->types[c7] == BLACK_PAWN)){
		*evalOpening -= TrappedBishop;
		*evalEndgame -= TrappedBishop;
	}
	if((board->types[h7] ==  WHITE_BISHOP && board->types[g6] == BLACK_PAWN) ||
	   (board->types[g8] == WHITE_BISHOP && board->types[f7] == BLACK_PAWN)){
		*evalOpening -= TrappedBishop;
		*evalEndgame -= TrappedBishop;
	}
	if((board->types[a2] ==  BLACK_BISHOP && board->types[b3] == WHITE_PAWN) ||
	   (board->types[b1] == BLACK_BISHOP && board->types[c2] == WHITE_PAWN)){
		*evalOpening += TrappedBishop;
		*evalEndgame += TrappedBishop;
	}
	if((board->types[h2] ==  BLACK_BISHOP && board->types[g3] == WHITE_PAWN) ||
	   (board->types[g1] == BLACK_BISHOP && board->types[f2] == WHITE_PAWN)){
		*evalOpening += TrappedBishop;
		*evalEndgame += TrappedBishop;
	}

	/* Trapped bishop (6th rank) */
	if (board->types[a6] == WHITE_BISHOP && board->types[b5] == BLACK_PAWN) {
      *evalOpening -= TrappedBishop / 2;
      *evalEndgame -= TrappedBishop / 2;
   }

   if (board->types[h6] == WHITE_BISHOP && board->types[g5] == BLACK_PAWN) {
      *evalOpening -= TrappedBishop / 2;
      *evalEndgame -= TrappedBishop / 2;
   }

   if (board->types[a3] == BLACK_BISHOP && board->types[b4] == WHITE_PAWN) {
      *evalOpening += TrappedBishop / 2;
      *evalEndgame += TrappedBishop / 2;
   }

   if (board->types[h3] == BLACK_BISHOP && board->types[g4] == WHITE_PAWN) {
      *evalOpening += TrappedBishop / 2;
      *evalEndgame += TrappedBishop / 2;
   }

	/* Blocked bishop */
	if(board->types[d2] == WHITE_PAWN && board->types[d3] != EMPTY && board->types[c1] == WHITE_BISHOP)
		*evalOpening -= BlockedBishop;
	if(board->types[e2] == WHITE_PAWN && board->types[e3] != EMPTY && board->types[f1] == WHITE_BISHOP)
		*evalOpening -= BlockedBishop;
	if(board->types[d7] == BLACK_PAWN && board->types[d6] != EMPTY && board->types[c8] == BLACK_BISHOP)
		*evalOpening += BlockedBishop;
	if(board->types[e7] == BLACK_PAWN && board->types[e6] != EMPTY && board->types[f8] == BLACK_BISHOP)
		*evalOpening += BlockedBishop;
	/* Blocked rook */
	if((board->types[c1] == WHITE_KING || board->types[b1] == WHITE_KING) &&
	   (board->types[a1] == WHITE_ROOK || board->types[a2] == WHITE_ROOK || board->types[b1] == WHITE_ROOK))
		*evalOpening -= BlockedRook;
	if((board->types[f1] == WHITE_KING || board->types[g1] == WHITE_KING) &&
	   (board->types[h1] == WHITE_ROOK || board->types[h2] == WHITE_ROOK || board->types[g1] == WHITE_ROOK))
		*evalOpening -= BlockedRook;
	if((board->types[c8] == BLACK_KING || board->types[b8] == BLACK_KING) &&
	   (board->types[a8] == BLACK_ROOK || board->types[a7] == BLACK_ROOK || board->types[b8] == BLACK_ROOK))
		*evalOpening += BlockedRook;
	if((board->types[f8] == BLACK_KING || board->types[g8] == BLACK_KING) &&
	   (board->types[h8] == BLACK_ROOK || board->types[h7] == BLACK_ROOK || board->types[g8] == BLACK_ROOK))
		*evalOpening += BlockedRook;
}

/**
 *
 */
void boardEvalPieces(const Board* board, const EvalInfo* minfo, gint* rOpening, gint* rEndgame){
	guint64 pieces, freeSquares, attack, mobAttack;
	gint player, opponent=0, from, pieceFile, pieceRank, kingFile, kingRank, distance;
	gint evalOpening[2], evalEndgame[2], mobility;

	freeSquares = board->freeSquares;
	for(player=WHITE; player<=BLACK; player++){
		
		evalOpening[player] = evalEndgame[player] = 0;
		opponent = !player;
		mobAttack = freeSquares | board->pieces[opponent];
		from = indexOfFirstBit64(&board->king[opponent]);
		kingFile = file(from);
		kingRank = rank(from);
	
		/* Knight */
		pieces = board->knights[player];
		while(pieces){
			from = getFirstBitAndClear64(&pieces);
			/* Mobility */
			mobility = -KnightUnit + numberOfBits64(attackKnight[from] & mobAttack);
			evalOpening[player] += mobility*KnightMobOpening;
			evalEndgame[player] += mobility*KnightMobEndgame;
			/* Outpost */
			if(player == WHITE){
				if(board->types[from-7] == WHITE_PAWN){
					evalOpening[player] += KnightOutpostMatrix[player][from];
				}
				if(board->types[from-9] == WHITE_PAWN){
					evalOpening[player] += KnightOutpostMatrix[player][from];
				}
			}else{
				if(board->types[from+7] == BLACK_PAWN){
					evalOpening[player] += KnightOutpostMatrix[player][from];
				}
				if(board->types[from+9] == BLACK_PAWN){
					evalOpening[player] += KnightOutpostMatrix[player][from];
				}
			}
		}
		/* Bishop */
		pieces = board->bishops[player];
		while(pieces){
			from = getFirstBitAndClear64(&pieces);
			/* Mobility */
			mobility = -BishopUnit + numberOfBits64(attackBishop(board,from) & mobAttack);
			evalOpening[player] += mobility * BishopMobOpening;
			evalEndgame[player] += mobility * BishopMobEndgame;
		}
		/* Rook */
		pieces = board->rooks[player];
		while(pieces){
			from = getFirstBitAndClear64(&pieces);
			/* Mobility */
			mobility = -RookUnit + numberOfBits64(attackRook(board,from) & mobAttack);
			evalOpening[player] += mobility * RookMobOpening;
			evalEndgame[player] += mobility * RookMobEndgame;
			/* Open file */
			pieceFile = file(from);
			evalOpening[player] -= RookOpenFileOpening / 2;
			evalEndgame[player] -= RookOpenFileEndgame / 2;
			attack = bitFile[pieceFile];
			if((attack & board->pawns[player]) == 0){ /* No friendly pawns*/
				evalOpening[player] += RookSemiOpenFileOpening;
				evalEndgame[player] += RookSemiOpenFileEndgame;
				if((attack & board->pawns[opponent]) == 0){ /* No enemy pawns */
					evalOpening[player] +=  RookOpenFileOpening - RookSemiOpenFileOpening;
					evalEndgame[player] +=  RookOpenFileEndgame - RookSemiOpenFileEndgame;
				}
				if(minfo->kingSafety[opponent]){
					distance = ABS(pieceFile-kingFile);
					if(distance <= 1){
						evalOpening[player] += RookSemiKingFileOpening;
						if(distance == 0){
							evalOpening[player] += 
								RookKingFileOpening - RookSemiKingFileOpening;
						}
					}
				}
			}
			/* 7th rank */
			pieceRank = rank(from);
			if(pieceRank == rank7[player]){
				/* Opponent pawns on 7th rank or the king on 8th rank */
				if(board->pawns[opponent] & bit7Rank[player] || 
					kingRank == rank8[player]){
						evalOpening[player] += Rook7thOpening;
						evalEndgame[player] += Rook7thEndgame;
				}
			}
		}
		/* Queen */
		pieces = board->queens[player];
		while(pieces){
			from = getFirstBitAndClear64(&pieces);
			/* Distance from king */
			pieceFile = file(from);
			pieceRank = rank(from);
			mobility = 10 -(ABS(kingRank-pieceRank)+ABS(kingFile-pieceFile));
			evalOpening[player] += mobility;
			evalEndgame[player] += mobility;
			/* 7th rank */
			if(pieceRank == rank7[player]){
				/* Opponent pawns on 7th rank or the king on rank 8 */
				if(board->pawns[opponent] & bit7Rank[player] || 
					kingRank == rank8[player]){
						evalOpening[player] += Queen7thOpening;
						evalEndgame[player] += Queen7thEndgame;
				}
			}
		}
	}
	*rOpening += (evalOpening[WHITE] - evalOpening[BLACK]);
	*rEndgame += (evalEndgame[WHITE] - evalEndgame[BLACK]);
}

guint boardShelterFile(const Board* board, guint file, guint rank, guint color) {
	guint distance, penalty;
	guint64 pieces;
	
	pieces = board->pawns[color] & bitFile[file];
	
	if(color == WHITE){
		distance = rank*8;
		pieces = (pieces >> distance)<<distance;
		if(pieces) distance = 7 - rank(indexOfFirstBit64(&pieces));
		else distance = 0;
	}
	else{
		distance = (7-rank)*8;
		pieces = (pieces << distance)>>distance;
		if(pieces) distance = rank(indexOfLastBit64(&pieces));
		else distance = 0;
	}

	penalty = 36 - distance * distance;
	return penalty;
}

guint boardStormFile(const Board * board, guint file, guint color) {
	guint distance, penalty = 0;
	guint64 pieces;

	pieces = board->pawns[!color] & bitFile[file];
	if(color == WHITE){
		if(pieces) distance = 7 - rank(indexOfFirstBit64(&pieces));
		else distance = 0;
	}
	else{
		if(pieces) distance = rank(indexOfLastBit64(&pieces));
		else distance = 0;
	}
	
	/*g_printf("DEBUG: distance%d\n",distance);*/
	
	switch (distance) {
		case 3:{ penalty = StormOpening * 1; break;}
		case 4:{ penalty = StormOpening * 3; break;}
		case 5:{ penalty = StormOpening * 6; break;}
	}
	return penalty;
}

gint boardShelterSquare(const Board* board, gint square, gint color){
	guint penalty=0, file, rank;

	file = file(square);
	rank = rank(square);

	penalty += boardShelterFile(board,file,rank,color) * 2;
	if (file != 0) penalty += boardShelterFile(board,file-1,rank,color);
	if (file != 7) penalty += boardShelterFile(board,file+1,rank,color);
	/*g_printf("DEBUG: shelter file %d\n",boardShelterFile(board,file,rank,color) * 2);*/

	if (penalty == 0) penalty = 11; /* Weak back rank */

	penalty += boardStormFile(board,file,color);
	/*g_printf("DEBUG: storm file %d\n", boardStormFile(board,file,color));*/
	if (file != 0){
			penalty += boardStormFile(board,file-1,color);
			/*g_printf("DEBUG: storm file -1 %d\n", boardStormFile(board,file-1,color));*/
	}
	if (file != 7){
			 penalty += boardStormFile(board,file+1,color);
			/*g_printf("DEBUG: storm file +1 %d\n", boardStormFile(board,file+1,color));*/
	}

	return penalty;
}

void boardEvalKing(const Board* board, const EvalInfo* eInfo, gint* rOpening, gint* rEndgame){
	guint kingIndex, kingFile, kingRank, penalty1,penalty2,tmp, player,opponent, attackNumber, attackTotal;
	guint64 pieces,kingSafeSquares1,kingSafeSquares2;
	gint evalOpening[2]={0,0},evalEndgame[2]={0,0}, penalty, from;
	
	/* White king */
	if(eInfo->kingSafety[WHITE]){
		/* King */
		kingIndex = indexOfFirstBit64(&board->king[WHITE]);
		penalty1 = boardShelterSquare(board,kingIndex,WHITE);
		/*g_printf("DEBUG: penalty %d\n",penalty1);*/

		/* Castling */
		penalty2 = penalty1;
		if((board->castling & WHITE_CASTLING_KING_SIDE) != 0){
			tmp = boardShelterSquare(board,g1,WHITE);
			if(tmp < penalty2) penalty2 = tmp;
		}
		if((board->castling & WHITE_CASTLING_QUEEN_SIDE) != 0){
			tmp = boardShelterSquare(board,b1,WHITE);
			if(tmp < penalty2) penalty2 = tmp;
		}
		penalty = (penalty1+penalty2)/2;
		evalOpening[WHITE] -= penalty;
	}

	/* Black king */
	if(eInfo->kingSafety[BLACK]){
		/* King */
		kingIndex = indexOfFirstBit64(&board->king[BLACK]);
		penalty1 = boardShelterSquare(board,kingIndex,BLACK);
		/*g_printf("DEBUG: penalty %d\n",penalty1);*/

		/* Castling */
		penalty2 = penalty1;
		if((board->castling & BLACK_CASTLING_KING_SIDE) != 0){
			tmp = boardShelterSquare(board,g8,BLACK);
			if(tmp <  penalty2) penalty2 = tmp;
		}
		if((board->castling & BLACK_CASTLING_QUEEN_SIDE) != 0){
			tmp = boardShelterSquare(board,b8,BLACK);
			if(tmp <  penalty2) penalty2 = tmp;
		}
		penalty = (penalty1+penalty2)/2;
		evalOpening[BLACK] -= penalty;
	}
	/*g_printf("DEBUG: safety %d %d\n",evalOpening[WHITE],evalOpening[BLACK]);*/

	/* King attack */
	for(player=WHITE; player <=BLACK; player++){
		opponent = !player;
		if(eInfo->kingSafety[player]){
			attackNumber = 0;
			attackTotal = 0;
			kingIndex = indexOfFirstBit64(&board->king[player]);
			kingFile = file(kingIndex);
			kingRank = rank(kingIndex);
			kingSafeSquares1 = maskKingSafeDistance1[kingIndex];
			kingSafeSquares2 = maskKingSafeDistance2[kingIndex];
			/* Knight */
			pieces = board->knights[opponent];
			while(pieces){
				from = getFirstBitAndClear64(&pieces);
				if(attackKnight[from] & kingSafeSquares1){
					attackNumber ++;
					attackTotal += KingAttackUnit[KNIGHT];
				}
			}
			/* Bishop */
			pieces = board->bishops[opponent];
			while(pieces){
				from = getFirstBitAndClear64(&pieces);
				if((maskBishop[from] & kingSafeSquares1) && (attackBishop(board,from) & kingSafeSquares2)){
					attackNumber ++;
					attackTotal += KingAttackUnit[BISHOP];
				}
			}
			/* Rooks */
			pieces = board->rooks[opponent];
			while(pieces){
				from = getFirstBitAndClear64(&pieces);
				if((maskRook[from] & kingSafeSquares1) && (attackRook(board,from) & kingSafeSquares2)){
					attackNumber ++;
					attackTotal += KingAttackUnit[ROOK];
				}
			}
			/* Queens */
			pieces = board->queens[opponent];
			while(pieces){
				from = getFirstBitAndClear64(&pieces);
				if((maskQueen[from] & kingSafeSquares1) && (attackQueen(board,from) & kingSafeSquares2)){
					attackNumber ++;
					attackTotal += KingAttackUnit[QUEEN];
				}
			}
			evalOpening[player] -= (attackTotal * KingAttackOpening * KingAttackWeight[attackNumber]) / 256;
			/*g_printf("DEBUG: attack %d\n",(attackTotal * KingAttackOpening * KingAttackWeight[attackNumber]) / 256);*/
		}
	}
	*rOpening += evalOpening[WHITE] - evalOpening[BLACK];
	*rEndgame += evalEndgame[WHITE] - evalEndgame[BLACK];
}

inline gint quad(gint y_min, gint y_max, gint x) {
   gint y;
   y = y_min + ((y_max - y_min) * Bonus[x] + 128) / 256;
   return y;
}

void boardEvaluatePawns(const Board* board, EvalInfo* eInfo, gint* rOpening, gint* rEndgame){
	guint64 pawns, t1,t2;
	gint player, opponent, square, pawnFile, pawnRank, evalOpening[2]={0,0}, evalEndgame[2]={0,0},n;
	gboolean backward, doubled, isolated, open, passed, candidate;

	for(player = WHITE; player <= BLACK; player++){
		opponent = !player;
		pawns = board->pawns[player];
		eInfo->passedPawn[player] = 0;
		while(pawns){
			square = getFirstBitAndClear64(&pawns);
			pawnFile = file(square);
			pawnRank = rank(square);
			backward = doubled = isolated = open = passed = candidate= FALSE;

			if(bitsLE((bitFile[pawnFile] & board->pawns[player]),pawnRank)){ /* OK */
				doubled = TRUE;
			}
			t1 = getBitFile(board->pawns[player],pawnFile-1)|
					getBitFile(board->pawns[player],pawnFile+1);
			t2 = getBitFile(board->pawns[WHITE] | board->pawns[BLACK],pawnFile);
			
			if(!t1){ /* OK */
				isolated = TRUE;
			}else{ /* OK */
				if(player == WHITE && !bitsLEEQ(t1,pawnRank)){ /* OK */
					backward = TRUE;
					if((t1 & bitRank[pawnRank+1]) != 0){ /* OK */
						if(((t2 & bitRank[pawnRank+1]) | /* OK */
							((getBitFile(board->pawns[opponent],pawnFile-1) | 
							  getBitFile(board->pawns[opponent],pawnFile+1)) & 
							 (bitRank[pawnRank+1] | bitRank[pawnRank+2]))) == 0){
							backward = FALSE;
						}
					}
					else if(pawnRank == 1 && ((t1&bitRank[pawnRank+2]) != 0)){ /* OK */
						if(((t2 & (bitRank[pawnRank+1] | bitRank[pawnRank+2])) | /*OK */
							((getBitFile(board->pawns[opponent],pawnFile-1) | 
							  getBitFile(board->pawns[opponent],pawnFile+1)) & 
							 (bitRank[pawnRank+1] | bitRank[pawnRank+2] | bitRank[pawnRank+3]))) == 0)
							backward = FALSE;
					}
				}
				if(player == BLACK && !bitsGTEQ(t1,pawnRank)){ /* OK */
					backward = TRUE;
					if((t1 & bitRank[pawnRank-1]) != 0){ /* OK */
						if(((t2 & bitRank[pawnRank-1]) | /* OK */
							((getBitFile(board->pawns[opponent],pawnFile-1) | 
							  getBitFile(board->pawns[opponent],pawnFile+1)) & 
							 (bitRank[pawnRank-1] | bitRank[pawnRank-2]))) == 0 ){
							backward = FALSE;
						}
					}
					else if(pawnRank == 6 && ((t1&bitRank[pawnRank-2]) != 0)){ /* OK */
						if(((t2 & (bitRank[pawnRank-1] | bitRank[pawnRank-2])) | /* OK */
							((getBitFile(board->pawns[opponent],pawnFile-1) | 
							  getBitFile(board->pawns[opponent],pawnFile+1)) & 
							 (bitRank[pawnRank-1] | bitRank[pawnRank-2] | bitRank[pawnRank-3]))) == 0)
							backward = FALSE;
					}
				}
			}
			
			gchar sqstr[3];
			squareToString(square, sqstr);

			if(player == WHITE && !bitsGT(t2, pawnRank)){ /* OK */
				open = TRUE;
				if(!bitsGT(getBitFile(board->pawns[opponent], pawnFile-1),pawnRank) && /* OK */
					!bitsGT(getBitFile(board->pawns[opponent],pawnFile+1),pawnRank)){
					passed = TRUE;
					eInfo->passedPawn[player] |= 1ull<<square;
				}
				else{
					n = 0;
					n += getNumberBitsLEEQ(getBitFile(board->pawns[player], pawnFile-1),pawnRank);
					n += getNumberBitsLEEQ(getBitFile(board->pawns[player], pawnFile+1),pawnRank);
					n -= getNumberBitsGT(getBitFile(board->pawns[opponent], pawnFile-1),pawnRank);
					n -= getNumberBitsGT(getBitFile(board->pawns[opponent], pawnFile+1),pawnRank);
					if(n>=0){ /* OK */
						n = 0;
						n += getNumberBitsEQ(getBitFile(board->pawns[player], pawnFile-1),pawnRank-1);
						n += getNumberBitsEQ(getBitFile(board->pawns[player], pawnFile+1),pawnRank-1);
						n -= getNumberBitsEQ(getBitFile(board->pawns[opponent], pawnFile-1),pawnRank+1);
						n -= getNumberBitsEQ(getBitFile(board->pawns[opponent], pawnFile+1),pawnRank+1);
						if(n >= 0)	candidate = TRUE;
					}
				}
			}
			if(player == BLACK && !bitsLE(t2, pawnRank)){ /* OK */
				open = TRUE;
				if(!bitsLE(getBitFile(board->pawns[opponent], pawnFile-1),pawnRank) && /* OK */
					!bitsLE(getBitFile(board->pawns[opponent],pawnFile+1),pawnRank)){
					passed = TRUE;
					eInfo->passedPawn[player] |= 1ull<<square;
				}else{
					n = 0;
					n += getNumberBitsGTEQ(getBitFile(board->pawns[player], pawnFile-1),pawnRank);
					n += getNumberBitsGTEQ(getBitFile(board->pawns[player], pawnFile+1),pawnRank);
					n -= getNumberBitsLE(getBitFile(board->pawns[opponent], pawnFile-1),pawnRank);
					n -= getNumberBitsLE(getBitFile(board->pawns[opponent], pawnFile+1),pawnRank);
					if(n >= 0){ /* OK */
						n = 0;
						n += getNumberBitsEQ(getBitFile(board->pawns[player], pawnFile-1),pawnRank+1);
						n += getNumberBitsEQ(getBitFile(board->pawns[player], pawnFile+1),pawnRank+1);
						n -= getNumberBitsEQ(getBitFile(board->pawns[opponent], pawnFile-1),pawnRank-1);
						n -= getNumberBitsEQ(getBitFile(board->pawns[opponent], pawnFile+1),pawnRank-1);
						if(n >= 0)	candidate = TRUE;
					}
				}
			}

			if(doubled){
            	evalOpening[player] -= DoubledOpening;
				evalEndgame[player] -= DoubledEndgame;
         	}

        	if (isolated){
            	if (open){
					evalOpening[player] -= IsolatedOpeningOpen;
               		evalEndgame[player] -= IsolatedEndgame;
            	}else{
					evalOpening[player] -= IsolatedOpening;
					evalEndgame[player] -= IsolatedEndgame;
            	}
         	}

			if (backward){
            	if (open){
               		evalOpening[player] -= BackwardOpeningOpen;
               		evalEndgame[player] -= BackwardEndgame;
            	}else{
               		evalOpening[player] -= BackwardOpening;
               		evalEndgame[player] -= BackwardEndgame;
            	}
			}

			if (candidate) {
				if(player == WHITE){
					evalOpening[player] += quad(CandidateOpeningMin,CandidateOpeningMax,pawnRank);
    	        	evalEndgame[player] += quad(CandidateEndgameMin,CandidateEndgameMax,pawnRank);
				}else{
					evalOpening[player] += quad(CandidateOpeningMin,CandidateOpeningMax,7-pawnRank);
    	        	evalEndgame[player] += quad(CandidateEndgameMin,CandidateEndgameMax,7-pawnRank);
				}
         	}
		}
	}
	*rOpening += evalOpening[WHITE] - evalOpening[BLACK];
	*rEndgame += evalEndgame[WHITE] - evalEndgame[BLACK];
}

inline gboolean boardUnstoppablePasser(const Board * board, gint square, gint player){
		gint pRank, oppKing, dist;

		pRank = rank(square);

		if(player == WHITE){ /* two squares up */
			oppKing =  indexOfFirstBit64(&board->king[BLACK]);
			if(bitsGT(bitFile[file(square)] & board->pieces[WHITE],pRank)) /* Clear promotion path? */
				return FALSE;
			if(pRank == 1){
				pRank ++;
				square += 8;
			}
			dist = distance(oppKing,square+(8*(7-pRank)));
			if(board->player == BLACK) dist --;
			if(dist > 7 - pRank) return TRUE;

		}
		else{
			oppKing =  indexOfFirstBit64(&board->king[WHITE]);
			if(bitsLE(bitFile[file(square)] & board->pieces[BLACK],pRank)) /* Clear promotion path? */
				return FALSE;
			if(pRank == 6){ /* two squares down */
				pRank --;
				square -= 8;
			}
			dist = distance(oppKing,square-(8*pRank));
			if(board->player == WHITE) dist --;
			if(dist > pRank) return TRUE;
		}
		return FALSE;
}

inline gboolean boardKingPasser(const Board* board, gint square, gint player) {
	gint king, pFile, pRank, promotion;

	king =  indexOfFirstBit64(&board->king[player]);
	pFile = file(square);
	pRank = rank(square);

	if(player == WHITE) promotion = king + (8*(7-rank(king)));
	else promotion = king - (8*rank(king));
	
	if (distance(king,promotion) <= 1 && distance(king,square) == 1 && 
		(file(king) != pFile  || (pFile != 0 && pFile != 7))){
			return TRUE;
	}
	return FALSE;
}

inline gboolean boardFreePasser(const Board* board, gint square, gint player) {
	gint pawn;
	if(player == WHITE) pawn = square + 8;
	else pawn = square - 8;
	
	if(board->types[pawn] != EMPTY) return FALSE;
	
	return TRUE;
}

inline gint boardPawnAttDist(gint square, gint king, gint player) {
	gint pawn;
	if(player == WHITE) pawn = square + 8;
	else pawn = square - 8;
	return distance(pawn,king);
}

inline gint boardPawnDefDist(gint square, gint king, gint player) {
	gint pawn;
	if(player == WHITE) pawn = square + 8;
	else pawn = square - 8;
	return distance(pawn,king);
}

void boardEvalPasser(const Board* board, const EvalInfo* eInfo, gint* rOpening, gint* rEndgame){
		gint evalOpening[2] = {0,0}, evalEndgame[2] = {0,0}, min, max, delta;
		gint square, king, oppKing, player, opponent, pRank;
		guint64 passedPawn;

		king = indexOfFirstBit64(&board->king[WHITE]);
		oppKing = indexOfFirstBit64(&board->king[BLACK]);
		

		/*gchar str[3];*/
		for(player = WHITE; player <= BLACK; player++){
			passedPawn = eInfo->passedPawn[player];
			opponent = !player;
			if(player==BLACK){
				delta = king;
				king = oppKing;
				oppKing = delta;
			}
			while(passedPawn){
				square = getFirstBitAndClear64(&passedPawn);
				if(player == WHITE) pRank = rank(square);
				else pRank = 7 - rank(square);
				evalOpening[player] += quad(PassedOpeningMin,PassedOpeningMax, pRank);
				
				min = PassedEndgameMin;
				max = PassedEndgameMax;
				delta = max - min;

				/*squareToString(square,str);	*/
				/*g_printf("DEBUG: unstoppable %s %d\n",str,boardUnstoppablePasser(board,square,player));*/
				/*g_printf("DEBUG: king passer %s %d\n",str,boardKingPasser(board,square,player));*/
				/*g_printf("DEBUG: free passer %s %d\n",str,boardFreePasser(board,square,player));*/

				if(eInfo->majorMaterial[opponent] == 0 && 
					(boardUnstoppablePasser(board,square,player) || boardKingPasser(board,square,player))){
						delta += UnstoppablePasser;
				}
				/*else if(boardFreePasser(board,square,player)){
						delta += FreePasser;
				}*/
				/*g_printf("DEBUG: pawn att %s %d\n",str, boardPawnAttDist(square,king,player));*/
				/*g_printf("DEBUG: pawn def %s %d\n",str, boardPawnDefDist(square,oppKing,player));*/
				delta -= boardPawnAttDist(square,king,player) * AttackerDistance;
				delta += boardPawnDefDist(square,oppKing,player) * DefenderDistance;

				/*g_printf("DEBUG: pawn %s %d %d %d\n",str, min,delta,pRank);*/
				evalEndgame[player] += min;
				if (delta > 0) evalEndgame[player] += quad(0,delta,pRank);
			}
		}
		
		*rOpening += (evalOpening[WHITE] - evalOpening[BLACK]);
		*rEndgame += (evalEndgame[WHITE] - evalEndgame[BLACK]);
}

gint boardEvaluate(const Board* board, gint alpha,gint beta){
	gint eval, phase, evalOpening, evalEndgame;
	EvalInfo eInfo;
	boardMaterialInfo(board, &eInfo);

	gint numberOfPieces = board->numberOfPieces[WHITE] + board->numberOfPieces[BLACK];
	if(numberOfPieces <= 4)	return boardEvaluateEndGame(board);

	/* Material */
	phase = eInfo.phase;
	evalOpening = eInfo.evalOpening;
	evalEndgame = eInfo.evalEndgame;
	/*g_print("DEBUG: EVAL1 %d %d\n",evalOpening,evalEndgame);*/

	/* PST */
	evalOpening += board->evalOpening;
	evalEndgame += board->evalEndgame;
	/*g_print("DEBUG: EVAL2 %d %d\n",evalOpening,evalEndgame);*/

	/* Pawns */
	boardEvaluatePawns(board, &eInfo, &evalOpening, &evalEndgame);
	/*g_print("DEBUG: EVAL3 %d %d\n",evalOpening,evalEndgame);*/

	/* Draw */

	/* Tempo */
	if(board->player == WHITE){
		evalOpening += 20;
		evalEndgame += 10;
	}
	else{
		evalOpening -= 20;
		evalEndgame -= 10;
	}

	/* Patterns */
	/*g_printf("DEBUG: EVAL4 %d %d\n",evalOpening,evalEndgame);*/
	boardEvalPatterns(board, &evalOpening, &evalEndgame);
	/*g_printf("DEBUG: EVAL5 %d %d\n",evalOpening,evalEndgame);*/
	
	/* Lazy cutoff ?? */

	/* Pieces */
	boardEvalPieces(board, &eInfo, &evalOpening, &evalEndgame);
	/*g_printf("DEBUG: EVAL6 %d %d\n",evalOpening,evalEndgame);*/

	/* King */
	boardEvalKing(board, &eInfo, &evalOpening, &evalEndgame);
	/*g_printf("DEBUG: EVAL7 %d %d\n",evalOpening,evalEndgame);*/

	/* Passer */
	boardEvalPasser(board, &eInfo, &evalOpening, &evalEndgame);
	/*g_printf("DEBUG: EVAL8 %d %d\n",evalOpening,evalEndgame);*/

	/*g_print("DEBUG: phase %d\n",phase);*/
	eval = ((evalOpening * (256 - phase)) + (evalEndgame * phase)) / 256;
 
	if(board->player == BLACK) eval = -eval;
	/*g_printf("DEBUG: EVAL %d\n",eval);*/
	return eval;
}
