/***************************************************************************
 *   Copyright (C) 2006 by Borko Boskovic                                  *
 *   borko.boskovic@uni-mb.si                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
#include <ctype.h>
#include <glib/gprintf.h>

#include "move.h"

inline void boardMovePiece(Board* board, guint piece, gint from, guint to){
	guint64 bbsquare;
	bbsquare = 1ull << from | 1ull << to;
	board->pieces[piece%2] ^= bbsquare;  /* Mona pohitritev ? */
	board->pieces[piece] ^= bbsquare;
	board->freeSquares ^= bbsquare;
	board->piecesRL90 ^= 1ull << rotateL90[from] | 1ull << rotateL90[to]; 
	board->piecesRL45 ^= 1ull << rotateL45[from] | 1ull << rotateL45[to];
	board->piecesRR45 ^= 1ull << rotateR45[from] | 1ull << rotateR45[to];
	board->key ^= squareKey[piece][from] ^ squareKey[piece][to];
	board->types[from] = EMPTY;
	board->types[to] = piece;
	board->evalOpening -= pst[piece][from][Opening];
	board->evalOpening += pst[piece][to][Opening];
	board->evalEndgame -= pst[piece][from][Endgame];
	board->evalEndgame += pst[piece][to][Endgame];
}

inline void boardWhitePawnHandler(Board* board, guint from, guint to, guint cPiece, Move move){
	gint distance;
	if(to < a8){
		distance = to - from;
		switch(distance){
			case 7: { if(cPiece == EMPTY) boardRemovePiece(board,from-1); break; }
			case 9: { if(cPiece == EMPTY) boardRemovePiece(board,from+1); break; }
			case 16:{ board->enPassant = to-8; board->key ^= board->enPassant; break; }
		};
	}
	else{
		boardRemovePiece(board,to);
		boardAddPiece(board,moveGetPromotion(move),to);
	}
}

inline void boardWhitePawnUndoHandler(Board* board, guint from, guint to, guint piece, guint cPiece){
	gint distance = to - from;
	switch(distance){
		case 7: { if(cPiece == EMPTY) boardAddPiece(board,BLACK_PAWN,from-1); break; }
		case 9: { if(cPiece == EMPTY) boardAddPiece(board,BLACK_PAWN,from+1); break; }
	};
}

inline void boardBlackPawnHandler(Board* board, guint from, guint to, guint cPiece, Move move){
	gint distance;
	if(to > h1){
		distance = from - to;
		switch(distance){
			case 7: { if(cPiece == EMPTY) boardRemovePiece(board,from+1); break; }
			case 9: { if(cPiece == EMPTY) boardRemovePiece(board,from-1); break; }
			case 16:{ board->enPassant = to+8; board->key ^= board->enPassant; break; }
		};
		       
	}
	else{
		boardRemovePiece(board,to);
		boardAddPiece(board,moveGetPromotion(move),to);
	}
}

inline void boardBlackPawnUndoHandler(Board* board, guint from, guint to, guint piece, guint cPiece){
	gint distance = from - to;
	switch(distance){
		case 7: { if(cPiece == EMPTY) boardAddPiece(board,WHITE_PAWN,from+1); break; }
		case 9: { if(cPiece == EMPTY) boardAddPiece(board,WHITE_PAWN,from-1); break; }
	};
}

inline void boardUpdateCastling(Board* board, guint square){
	switch(square){
		case h1:{
			if(board->castling && WHITE_CASTLING_KING_SIDE)
				board->castling &= ~WHITE_CASTLING_KING_SIDE;
			break;
		}
		case a1:{
			if(board->castling && WHITE_CASTLING_QUEEN_SIDE)
				board->castling &= ~WHITE_CASTLING_QUEEN_SIDE;
			break;
		}
		case h8:{
			if(board->castling && BLACK_CASTLING_KING_SIDE)
				board->castling &= ~BLACK_CASTLING_KING_SIDE;
			break;
		}
		case a8:{
			if(board->castling && BLACK_CASTLING_QUEEN_SIDE)
				board->castling &= ~BLACK_CASTLING_QUEEN_SIDE;
			break;
		}
		
	};
}

inline gboolean whiteCanCastlingKingSide(Board* board){
	if((attackRook(board,e1)|attackRook(board,f1)|attackRook(board,g1) )&
		(board->rooks[BLACK]|board->queens[BLACK])) return FALSE;
	if((attackBishop(board,e1)|attackBishop(board,f1)|attackBishop(board,g1))&
		(board->bishops[BLACK]|board->queens[BLACK]) ) return FALSE;
	if(16309248ull&board->knights[BLACK]) return FALSE;
	if(63488ull&board->pawns[BLACK]) return FALSE;
	if(63736ull&board->king[BLACK]) return FALSE;
	return TRUE;
}

inline gboolean whiteCanCastlingQueenSide(Board* board){
	if((attackRook(board,c1)|attackRook(board,d1)|attackRook(board,e1) )&
		(board->rooks[BLACK]|board->queens[BLACK])) return FALSE;
	if((attackBishop(board,c1)|attackBishop(board,d1)|attackBishop(board,e1))&
		(board->bishops[BLACK]|board->queens[BLACK])) return FALSE;
	if(4093696ull&board->knights[BLACK]) return FALSE;
	if(15872ull&board->pawns[BLACK]) return FALSE;
	if(15934ull&board->king[BLACK]) return FALSE;
	return TRUE;
}

inline gboolean blackCanCastlingKingSide(Board* board){
	if((attackRook(board,e8)|attackRook(board,f8)|attackRook(board,g8))&
		(board->rooks[WHITE]|board->queens[WHITE])) return FALSE;
	if((attackBishop(board,e8)|attackBishop(board,f8)|attackBishop(board,g8))&
		(board->bishops[WHITE]|board->queens[WHITE])) return FALSE;
	if(62197173760032768ull&board->knights[WHITE]) return FALSE;
	if(69805794224242688ull&board->pawns[WHITE]) return FALSE;
	if(17940089115630370816ull&board->king[WHITE]) return FALSE;
	return TRUE;
}

inline gboolean blackCanCastlingQueenSide(Board* board){
	if((attackRook(board,c8)|attackRook(board,d8)|attackRook(board,e8))&
		(board->rooks[WHITE]|board->queens[WHITE])) return FALSE;
	if( (attackBishop(board,c8)|attackBishop(board,d8)|attackBishop(board,e8))&
		(board->bishops[WHITE]|board->queens[WHITE]) ) return FALSE;
	if(33563691949490176ull&board->knights[WHITE]) return FALSE;
	if(17451448556060672ull&board->pawns[WHITE]) return FALSE;
	if(4485022278907592704ull&board->king[WHITE]) return FALSE;
	return TRUE;
}

inline gboolean boardWhiteKingHandler(Board* board, guint from, guint to){
	gint distance;
	distance = to - from;
	if(distance == 2){
		if(!whiteCanCastlingKingSide(board)) return FALSE;
		boardMovePiece(board,WHITE_ROOK,h1,f1);
	}
	else if (distance == -2){
		if(!whiteCanCastlingQueenSide(board)) return FALSE;
		boardMovePiece(board,WHITE_ROOK,a1,d1);
	}
	board->castling &= ~WHITE_CASTLING;
	return TRUE;
}

inline void boardWhiteKingUndoHandler(Board* board, guint from, guint to){
	gint distance;
	distance = to - from;
	if(distance == 2) boardMovePiece(board,WHITE_ROOK,f1,h1);
	else if (distance == -2) boardMovePiece(board,WHITE_ROOK,d1,a1);
}

inline gboolean boardBlackKingHandler(Board* board, guint from, guint to){
	gint distance;
	distance = to - from;
	if(distance == 2){
		if(!blackCanCastlingKingSide(board)) return FALSE;
		boardMovePiece(board,BLACK_ROOK,h8,f8);
	}
	else if (distance == -2){
		if(!blackCanCastlingQueenSide(board)) return FALSE;
		boardMovePiece(board,BLACK_ROOK,a8,d8);
	}
	board->castling &= ~BLACK_CASTLING;
	return TRUE;
}

inline void boardBlackKingUndoHandler(Board* board, guint from, guint to){
	gint distance;
	distance = to - from;
	if(distance == 2) boardMovePiece(board,BLACK_ROOK,f8,h8);
	else if (distance == -2)boardMovePiece(board,BLACK_ROOK,d8,a8);
}

gboolean boardMakeMove(Board* board, Move move){
	guint from, to, piece, cPiece;
	from = moveGetFrom(move);
	to = moveGetTo(move);
	piece = board->types[from];
	cPiece = board->types[to];
	#ifndef G_DISABLE_ASSERT
	if(piece == EMPTY){
		boardPrint(board);
		g_printf("Move: ");
		movePrint(move);
		fflush(stdout);
		g_error("Error: piece is empty (make move)!\n");
	}
	#endif
	board->undoCapturePiece[board->ply] = cPiece;
	board->undoCastling[board->ply] = board->castling;
	board->undoEnPassant[board->ply] = board->enPassant;
	board->undoCheck[board->ply] = board->check;
	board->historyOfKeys[board->ply] = board->key;
	board->key ^= board->enPassant;
	board->enPassant = 0;
	if(cPiece != EMPTY){ 
		boardRemovePiece(board,to); 
		boardUpdateCastling(board,to);
		board->numberOfStaticMoves[board->ply+1] = 0;
	}
	else
		board->numberOfStaticMoves[board->ply+1] = board->numberOfStaticMoves[board->ply]+1;
	boardMovePiece(board,piece,from,to);
	switch(piece){
		case WHITE_PAWN:{ 
			boardWhitePawnHandler(board,from,to,cPiece,move);
			board->numberOfStaticMoves[board->ply+1] = 0;
			break; 
		}
		case BLACK_PAWN:{ 
			boardBlackPawnHandler(board,from,to,cPiece,move);
			board->numberOfStaticMoves[board->ply+1] = 0;
			break; 
		}
		case WHITE_ROOK:{ boardUpdateCastling(board,from); break; }
		case BLACK_ROOK:{ boardUpdateCastling(board,from); break; }
		case WHITE_KING:{ 
			if(!boardWhiteKingHandler(board,from,to)){
				boardMovePiece(board,piece,to,from);
				cPiece = board->undoCapturePiece[board->ply];
				board->castling = board->undoCastling[board->ply];
				board->enPassant = board->undoEnPassant[board->ply];
				board->check = board->undoCheck[board->ply];
				board->key = board->historyOfKeys[board->ply];
				return FALSE;
			}
			break;
		}
		case BLACK_KING:{ 
			if(!boardBlackKingHandler(board,from,to)){
				boardMovePiece(board,piece,to,from);
				cPiece = board->undoCapturePiece[board->ply];
				board->castling = board->undoCastling[board->ply];
				board->enPassant = board->undoEnPassant[board->ply];
				board->check = board->undoCheck[board->ply];
				board->key = board->historyOfKeys[board->ply];
				return FALSE;
			} 
			break; 
		}
	};
	changePlayer(board);
	board->historyOfMoves[board->ply] = move;
	board->ply ++;
	if(isIllegal(board)){
		boardUnmakeMove(board);
		return FALSE;
	}
	board->check = isInCheck(board);
	#ifndef G_DISABLE_ASSERT
	boardValidate(board);
	#endif
	return TRUE;
}

Move boardUnmakeMove(Board* board){
	guint from, to, piece, cPiece;
	board->ply --;
	Move move = board->historyOfMoves[board->ply];
	uChangePlayer(board);
	from = moveGetFrom(move);
	to = moveGetTo(move);
	piece = board->types[to];
	cPiece = board->undoCapturePiece[board->ply];
	#ifndef G_DISABLE_ASSERT
	if(piece == EMPTY){
		boardPrint(board);
		g_printf("Move: ");
		movePrint(move);
		fflush(stdout);
		g_error("Error: piece is empty (unmake move)!\n");
	}
	#endif
	if(moveGetPromotion(move) != EMPTY){
		boardRemovePiece(board,to);
		boardAddPiece(board,PAWN + board->player,from);
		
	}else{
		boardMovePiece(board,piece,to,from);
	}
	if(cPiece != EMPTY) boardAddPiece(board,cPiece,to); 
	switch(piece){
		case WHITE_PAWN:{ boardWhitePawnUndoHandler(board,from,to,piece,cPiece); break; }
		case BLACK_PAWN:{ boardBlackPawnUndoHandler(board,from,to,piece,cPiece); break; }
		case WHITE_KING:{ boardWhiteKingUndoHandler(board,from,to); break; }
		case BLACK_KING:{ boardBlackKingUndoHandler(board,from,to); break; }
	};
	board->castling = board->undoCastling[board->ply];
	board->enPassant = board->undoEnPassant[board->ply];
	board->check = board->undoCheck[board->ply];
	board->key = board->historyOfKeys[board->ply];
	#ifndef G_DISABLE_ASSERT
	boardValidate(board);
	#endif
	return move;
}

void boardMakeNullMove(Board* board){
	board->historyOfKeys[board->ply] = board->key;	
	board->key ^= board->enPassant;
	board->undoEnPassant[board->ply] = board->enPassant;
	board->undoCastling[board->ply] = board->castling;
	board->enPassant = 0;
	changePlayer(board);	
	board->historyOfMoves[board->ply] = NULL_MOVE;
	board->ply ++;
}

void boardUnmakeNullMove(Board* board){
	board->ply--;
	uChangePlayer(board);
	board->enPassant = board->undoEnPassant[board->ply];
	board->castling = board->undoCastling[board->ply];
	board->key = board->historyOfKeys[board->ply];
}

void boardMakeMovesFromString(Board* board, gchar moves[]){
	guint index=0, gindex;
	Moves gmoves;
	gchar smove[6], sgmove[6];
	gboolean found;
	gboolean last = FALSE;
	while(!last){
		boardGenerateMoves(board, &gmoves);
		smove[0] = '\0';
		sscanf(moves,"%s",smove);
		if(moves[strlen(smove)] == '\0' || moves[strlen(smove)] == '\n') last = TRUE;
		if(strlen(smove) == 0) break;
		moves = g_strstr_len(moves, strlen(moves), smove) + strlen(smove);
		found = FALSE;
		for(gindex = 0; gindex<gmoves.size; gindex++){
			moveToString(gmoves.move[gindex], sgmove);
			if(!g_ascii_strcasecmp(smove, sgmove)){
				found = TRUE;
				if(!boardMakeMove(board,gmoves.move[gindex])){
					boardPrint(board);
					fflush(stdout);
					g_error("Wrong move:%s\n",smove);
				}
				break;
			}		
		}
		if(!found){
			boardPrint(board);
			fflush(stdout);
		    g_error("Can not make move (%s)!\n",smove);
		}
		index++;
	}
}

gboolean boardMakeNextMove(Board* board, Moves* moves){
	gint index,value;
	Move move;
	while(moves->next < moves->size){
		if(moves->next >= moves->sorted){
			for(index = moves->size-1; index>moves->next; index--){
				if(moves->value[index-1] < moves->value[index]){
					value = moves->value[index];
					moves->value[index] = moves->value[index-1];
					moves->value[index-1] = value;
			
					move = moves->move[index];
					moves->move[index] = moves->move[index-1];
					moves->move[index-1] = move;
					moves->sorted = index-1;
				}
			}
		}
		if(boardMakeMove(board,moves->move[moves->next])){
			moves->next ++;
			return TRUE;
		}
		moves->next ++;
	}
	return FALSE;
}
