/***************************************************************************
 *   Copyright (C) 2006 by Borko Boskovic                                  *
 *   borko.boskovic@uni-mb.si                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
/**
 * \file search.h
 * Search functions. Search begins with searchStart(Board*,gint,gint)
 * function and is stopped with searchStop() function.
 * Auxilary functions for collection of statistics are:
 * - searchReset(): resets statistics of the search,
 * - searchSetDepth(guint): sets search depth currently being searched and
 * - searchSetNodes(guint): sets number of nodes searched in current search.
 *
 * Definitions file contains additional local functions which implement the
 * search. These are:
 * - alphaBeta(Board*,gint,gint,gint,Move[]): non-root alpha-beta search,
 * - quiescent(Board*,gint,gint): being called for evaluation and dynamic moves, 
 * - boardEvaluate(Board*,gint,gint): statically evaluates the given position.
 */
#ifndef SEARCH_H
#define SEARCH_H

#include "board.h"
#include "transpositionTable.h"

/** Resets the search statistics and sets searchStopConditions to
 * stop.
 */
void searchReset();
/** Starts the search thread, first stops any search currently active.
 * The new thread that is created starts with the search(Board*)
 * function. The search function calls alphaBeta(Board*,gint,gint,gint,Move[])
 * function and this one calls quiescent(Board*,gint,gint) function 
 * at the lower levels and finally this one calls the
 * boardEvaluate(Board*,gint,gint) function.
 */
void searchStart(Board* board, TranspositionTable* trTable);
/** Sets depth to search.
 */
void searchSetDepth(guint depth);
/** Sets statistics for nodes searched. 
 */
void searchSetNodes(guint nodes);
/** Stops search.
 */
void searchStop();

void searchSetMultiPV(guint nmultipv);

#endif
